import importlib
import importlib.util

if importlib.util.find_spec("pydantic") is None:
    raise ImportError("Pydantic is no longer a required dependency. Please install it.")

from . import v1
from .v2 import Model, MutableModel

__all__ = ["Model", "MutableModel", "v1"]



import asyncio
from time import time
from typing import Any
from collections.abc import Callable
from collections.abc import Coroutine
from typing import Generic
from typing import TypeVar

from gyver.attrs import call_init
from gyver.attrs import mutable

from gyver.exc import ErrorGroup

from .resource import Resource

T = TypeVar("T")


FactoryType = Callable[[], Coroutine[Any, Any, T]]
ReleaserType = Callable[[T], Coroutine[Any, Any, None]]


@mutable
class AsyncPool(Generic[T]):
    resources: asyncio.Queue[Resource[T]]
    factory: FactoryType[T]
    releaser: ReleaserType[T]
    _pool_recycle: float
    _pool_size: int
    _available: int
    _available_semaphore: asyncio.Lock

    def __init__(
        self,
        factory: FactoryType[T],
        releaser: ReleaserType[T],
        queue_class: type[asyncio.Queue[T]] = asyncio.LifoQueue,
        pool_size: int = 10,
        pool_recycle: float = 3600,
    ):
        """
        Initialize the AsyncPool.

        :param factory: A coroutine function that creates a new resource.
        :param releaser: A coroutine function that releases a resource.
        :param queue_class: The class to use for the resources queue. Defaults to asyncio.LifoQueue.
        :param pool_size: The maximum size of the resource pool. Defaults to 10.
        :param pool_recycle: The time in seconds after which a resource is considered expired and should be recycled. Defaults to 3600 (1 hour).
        """
        call_init(
            self,
            resources=queue_class(pool_size),
            factory=factory,
            releaser=releaser,
            pool_recycle=pool_recycle,
            pool_size=pool_size,
            available=pool_size,
            available_semaphore=asyncio.Lock(),
        )

    async def _initialize_resource(self) -> Resource[T]:
        """Creates a resource from the factory and returns it wrapped in the Resource object
        :return: The Resource wrapper
        """
        resource = await self.factory()
        return Resource.from_now(resource)

    async def _maybe_recycle(
        self, resource: Resource[T], current_ts: float | None = None
    ) -> T:
        """Returns resource from resource object or
        releases the current resource and returns a new one if
        resource is expired.

        :param resource: The resource wrapper to be "maybe" recycled.
        :param current_ts: timestamp, defaults to `time.time()`

        :return: The resource acquired.
        """
        current_ts = current_ts or time()
        if (
            self._pool_recycle > 0
            and resource.last_usage + self._pool_recycle <= current_ts
        ):
            await self.releaser(resource.get())
            resource = await self._initialize_resource()
        return resource.get()

    async def acquire(self) -> T:
        """
        Acquires a resource from the pool, waiting until available.

        :return: T
        """
        acquired = await self._decrease_available()
        if not acquired:
            resource = await self.resources.get()
        else:
            try:
                resource = self.resources.get_nowait()
            except asyncio.QueueEmpty:
                resource = await self._initialize_resource()
        return await self._maybe_recycle(resource)

    async def release(self, resource: T) -> None:
        """Puts resource back in queue and increases the
        amount of resources available to be acquired

        :param resource: The resource to be put.
        :return: None
        """
        self.resources.put_nowait(Resource.from_resource(resource))
        await self._increase_available()

    async def prefill(self, count: int | None = None) -> None:
        """
        Prefills the queue by the amount passed.

        :param count: The amount of resources to initialize, up to the
            `pool_size`. Defaults to the pool_size if None.
        """
        count = count or self._pool_size
        count = min(count, self._pool_size)

        while self.resources.qsize() < count and await self._decrease_available():
            resource = await self._initialize_resource()
            self.resources.put_nowait(resource)

    async def dispose(self) -> None:
        """
        Remove and release all items of from the queue.

        :return: None
        :raises: ErrorGroup if any error happened while closing
        """
        errors = []
        while not self.resources.empty():
            resource = await self.resources.get()
            try:
                await self.releaser(resource.get())
            except Exception as e:
                errors.append(e)
            await self._decrease_available()
        if errors:
            raise ErrorGroup("Could not kill all resources", *errors)

    async def _decrease_available(self):
        """
        Decrease the count of available resources.

        This method should not be called directly. It is used internally
        by the `acquire` and `dispose` methods to keep track of the number
        of available resources.
        """
        async with self._available_semaphore:
            if self._available == 0:
                return False
            self._available -= 1
            return True

    async def _increase_available(self):
        """
        Increase the count of available resources.

        This method should not be called directly. It is used internally by the
        `release` and `prefill` methods to keep track of the number
        of available resources.
        """
        async with self._available_semaphore:
            if self._available == self._pool_size:
                return False
            self._available += 1
            return True



from urllib.parse import quote

from typing_extensions import Self

from gyver.attrs import mutable
from gyver.url.encode import Encodable
from gyver.url.utils import utf8


@mutable(eq=False)
class Netloc(Encodable):
    """Represents the network location portion of a URL.

    Attributes:
        username (Optional[str]): The username for authentication.
        password (Optional[str]): The password for authentication.
        host (str): The host name or IP address.
        port (Optional[int]): The port number.
    """

    username: str | None
    password: str | None
    host: str
    port: int | None

    def __init__(self, netloc: str) -> None:
        """
        Initialize the Netloc object.

        Args:
            netloc (str): The network location string.
        """
        self.username: str | None = None
        self.password: str | None = None
        self.load(netloc)

    def load(self, netloc: str):
        """
        Load the network location from the given netloc string.

        Args:
            netloc (str): The network location string.
        """
        userinfo, _, host = netloc.partition("@")
        port = None
        if host:
            username, _, password = userinfo.partition(":")
            self.username = username
            self.password = password
        else:
            host = userinfo
        if ":" in host:
            host, _, port = host.partition(":")
        self.host = host
        port = int(port) if port else None
        self.port = port

    def encode(self) -> str:
        """
        Encode the network location into a string.

        Returns:
            str: The encoded network location string.
        """
        netloc = ""
        if self.username:
            netloc = f"{quote(utf8(self.username))}"
            if self.password:
                netloc += f":{quote(utf8(self.password))}"
            netloc += "@"
        netloc += self.host
        if self.port:
            netloc += f":{self.port}"
        return netloc

    def parse(self, netloc: str):
        """
        Parse the given netloc string and populate the properties of the Netloc object.

        Args:
            netloc (str): The netloc string to parse.
        """
        if "@" in netloc:
            userinfo, host = netloc.split("@", 1)
            if ":" in userinfo:
                self.username, self.password = userinfo.split(":", 1)
            else:
                self.username = userinfo
            self.host = host
        elif ":" in netloc:
            self.host, port = netloc.split(":", 1)
            if port:
                self.port = int(port)
        else:
            self.host = netloc

    def set(
        self,
        host: str | None = None,
        username: str | None = None,
        password: str | None = None,
        port: int | None = None,
    ) -> Self:
        """
        Set the network location properties.

        Args:
            host (Optional[str], optional): The host name or IP address.
            username (Optional[str], optional): The username for authentication.
            password (Optional[str], optional): The password for authentication.
            port (Optional[int], optional): The port number.
        """
        self.host = host or self.host
        self.username = username or self.username
        self.password = password or self.password
        self.port = port or self.port
        return self

    def merge(self, netloc: "Netloc") -> "Netloc":
        """
        Merge the properties of the given `Netloc` object with the current object.

        Args:
            netloc (Netloc): The `Netloc` object to merge.

        Returns:
            Netloc: The merged `Netloc` object.
        """
        host = netloc.host or self.host
        username = netloc.username or self.username
        password = netloc.password or self.password
        port = netloc.port or self.port
        return self.from_args(host, username, password, port or None)

    @classmethod
    def from_args(
        cls,
        host: str,
        username: str | None = None,
        password: str | None = None,
        port: int | None = None,
    ) -> Self:
        """
        Create a new `Netloc` object from individual arguments.

        Args:
            host (str): The host name or IP address.
            username (Optional[str], optional): The username for authentication.
            password (Optional[str], optional): The password for authentication.
            port (Optional[int], optional): The port number.

        Returns:
            Netloc: The created `Netloc` object.
        """
        netloc = cls("")
        return netloc.set(host, username, password, port)

    def copy(self) -> "Netloc":
        netloc = object.__new__(type(self))
        netloc.host = self.host
        netloc.username = self.username
        netloc.password = self.password
        netloc.port = self.port
        return netloc



import typing

from gyver.context.typedef import T


class Adapter(typing.Protocol[T]):
    def is_closed(self, client: T) -> bool:
        """Returns whether the client state is closed or released."""
        ...

    def release(self, client: T) -> None:
        """Closes or releases the client."""
        ...

    def new(self) -> T:
        """Creates a new client."""
        ...


class AsyncAdapter(typing.Protocol[T]):
    """Represents an Adapter to a Specific Service."""

    async def is_closed(self, client: T) -> bool:
        """Returns whether the client state is closed or released."""
        ...

    async def release(self, client: T) -> None:
        """Closes or releases the client."""
        ...

    async def new(self) -> T:
        """Creates a new client."""
        ...


class AtomicAdapter(Adapter[T], typing.Protocol[T]):
    """Represents an Adapter to a Specific Service that can perform atomic operations."""

    def begin(self, client: T) -> None:
        """Starts an atomic operation."""
        ...

    def commit(self, client: T) -> None:
        """Commits an atomic operation."""
        ...

    def rollback(self, client: T) -> None:
        """Rolls back an atomic operation."""
        ...

    def in_atomic(self, client: T) -> bool:
        """Returns whether the client is currently in an atomic operation."""
        ...


class AtomicAsyncAdapter(AsyncAdapter[T], typing.Protocol[T]):
    """Represents an Adapter to a Specific Service that can perform atomic operations."""

    async def begin(self, client: T) -> None:
        """Starts an atomic operation."""
        ...

    async def commit(self, client: T) -> None:
        """Commits an atomic operation."""
        ...

    async def rollback(self, client: T) -> None:
        """Rolls back an atomic operation."""
        ...

    async def in_atomic(self, client: T) -> bool:
        """Returns whether the client is currently in an atomic operation."""
        ...



from config import MISSING, Config, EnvMapping, comma_separated
from config.enums import Env
from config.envconfig import DotFile, EnvConfig
from config.interface import ConfigLike
from config.utils import boolean_cast, joined_cast, valid_path, with_rule

from .adapter import AdapterConfigFactory
from .adapter.helpers import AttributeLoader, attribute, load, parametrize
from .adapter.mark import as_config, mark

__all__ = [
    "Config",
    "EnvMapping",
    "MISSING",
    "Env",
    "DotFile",
    "EnvConfig",
    "ConfigLike",
    "boolean_cast",
    "comma_separated",
    "valid_path",
    "joined_cast",
    "with_rule",
    "AdapterConfigFactory",
    "mark",
    "as_config",
    "load",
    "parametrize",
    "AttributeLoader",
    "attribute",
]



from typing import TYPE_CHECKING
from typing import Any
from collections.abc import Generator
from typing import Protocol
from collections.abc import Sequence
from typing import TypeVar

from config import MISSING

if TYPE_CHECKING:
    from dataclasses import Field

    from attrs import Attribute
    from gyver.attrs.field import Field as GField

    from .pydantic import FieldWrapper

T_co = TypeVar("T_co", "FieldWrapper", "Attribute", "Field", "GField", covariant=True)


class FieldResolverStrategy(Protocol[T_co]):
    """
    A protocol for resolving the properties of a field, regardless of
    its implementation.

    The `FieldResolver` protocol is defined to be generic over `T`,
    which is restricted to be one of three possible types:
    - `ModelField` from the `pydantic` library
    - `Attribute` from the `attrs` library
    - `Field` from the `dataclasses` library

    These types indicate the different ways that a field can be defined in Python.
    """

    def __init__(self, field: T_co) -> None:
        """
        Initialize a new `FieldResolver` instance.

        :param field: An instance of one of the three types `T` can be.
        """
        ...

    def names(self) -> Sequence[str]:
        """
        Get the names by which the field can be accessed.

        :return: A sequence of strings, representing the names.
        """
        ...

    def init_name(self) -> str:
        """
        Get the name by which the field is expected in init.

        :return: A sequence of strings, representing the names.
        """
        ...

    def cast(self) -> type:
        """
        Get the type of the field.

        :return: The type of the field.
        """
        ...

    def default(self) -> Any | type[MISSING]:
        """
        Get the default value of the field.

        :return: The default value of the field,
        or the special value `MISSING` (from the `config` module)
        if no default is defined.
        """
        ...

    @staticmethod
    def iterfield(config_class: type) -> Generator[T_co, Any, Any]: ...



from dataclasses import MISSING as dc_MISSING
from dataclasses import Field
from dataclasses import fields
from typing import Any
from collections.abc import Generator
from collections.abc import Sequence

from config import MISSING
from gyver.attrs import define

from gyver.config.adapter.interface import FieldResolverStrategy


@define
class DataclassResolverStrategy(FieldResolverStrategy[Field]):
    field: Field

    def cast(self) -> type:
        return self.field.type

    def names(self) -> Sequence[str]:
        return (self.field.name,)

    def init_name(self) -> str:
        return self.field.name

    def default(self) -> Any | type[MISSING]:
        if self.field.default not in (None, Ellipsis, dc_MISSING):
            return self.field.default
        return (
            MISSING
            if self.field.default_factory in (None, dc_MISSING)
            else self.field.default_factory
        )

    @staticmethod
    def iterfield(config_class: type) -> Generator[Field, Any, Any]:
        yield from fields(config_class)



import sys
from io import SEEK_SET, BytesIO
from typing import Generic

import cchardet
from typing_extensions import Self, TypeVar

from gyver.attrs import define

T = TypeVar("T")


@define
class VirtualPath(Generic[T]):
    name: str
    contents: T


class File(VirtualPath[BytesIO]):
    def get_encoding(self):
        val = cchardet.detect(self.getvalue())
        self.seek(SEEK_SET)
        return val["encoding"] or sys.getdefaultencoding()

    @classmethod
    def new(cls, name: str) -> Self:
        return cls(name, BytesIO())

    def read(self, size: int = -1) -> bytes:
        return self.contents.read(size)

    def readline(self, size: int = -1) -> bytes:
        return self.contents.readline(size)

    def readlines(self, hint: int = -1) -> list[bytes]:
        return self.contents.readlines(hint)

    def write(self, content: bytes) -> int:
        return self.contents.write(content)

    def writelines(self, lines: list[bytes]):
        self.contents.writelines(lines)

    def seek(self, offset: int, whence: int = SEEK_SET) -> int:
        return self.contents.seek(offset, whence)

    def tell(self) -> int:
        return self.contents.tell()

    def flush(self):
        self.contents.flush()

    def close(self):
        self.contents.close()

    def __enter__(self) -> "File":
        return self

    def __exit__(self, *_):
        self.close()

    def getvalue(self):
        return self.contents.getvalue()

    def __parse_dict__(self, by_alias: bool):
        return {"name": self.name, "contents": self.contents}


@define
class TextFile(File):
    encoding: str

    @classmethod
    def new(cls, name: str):
        return cls(name, BytesIO(), sys.getdefaultencoding())

    def read(self, size: int = -1) -> str:
        return self.contents.read(size).decode(self.encoding)

    def readline(self, size: int = -1) -> str:
        return self.contents.readline(size).decode(self.encoding)

    def readlines(self, hint: int = -1) -> list[str]:
        return [line.decode(self.encoding) for line in self.contents.readlines(hint)]

    def write(self, text: str) -> int:
        return self.contents.write(text.encode(self.encoding))

    def writelines(self, lines: list[str]):
        encoded_lines = [line.encode(self.encoding) for line in lines]
        self.contents.writelines(encoded_lines)

    def getvalue(self) -> str:
        return self.contents.getvalue().decode(self.encoding)


class Folder(VirtualPath[dict[str, VirtualPath]]):
    @classmethod
    def new(cls, name: str) -> Self:
        return cls(name, {})

    def __parse_dict__(self, by_alias: bool):
        return {
            "name": self.name,
            "contents": {
                key: value.__parse_dict__(by_alias)
                for key, value in self.contents.items()
            },
        }



import re
from collections.abc import Callable
from typing import Concatenate, TypeVar

from typing_extensions import ParamSpec

T = TypeVar("T")
R = TypeVar("R")
P = ParamSpec("P")


def try_cast(
    val: T, cast: Callable[Concatenate[T, P], R], *args: P.args, **kwargs: P.kwargs
) -> T | R:
    """
    Try to cast a value to a different type using a provided casting function.

    :param val: The value to be cast.
    :param cast: The casting function.
    :param args: Positional arguments to be passed to the casting function.
    :param kwargs: Keyword arguments to be passed to the casting function.
    :return: The casted value if successful, otherwise the original value.
    """
    try:
        return cast(val, *args, **kwargs)
    except Exception:
        return val


def utf8(val: str):
    """
    Encode a string using UTF-8.

    :param val: The string to be encoded.
    :return: The encoded string.
    """
    return try_cast(val, str.encode, "utf-8")


PERCENT_REGEX = r"\%[a-fA-F\d][a-fA-F\d]"


is_valid_encoded_path = re.compile(
    r"^([\w{}]|({}))*$".format(re.escape("-.~:@!$&'()*+,;="), PERCENT_REGEX)
)
